﻿// --------------------------------------------------
// Official site, including explanation on how to set up everything:
// http://fsharp.org/
// Or try it directly from your browser:
// http://www.tryfsharp.org/


// --------------------------------------------------
// Interactivity.
// Basic examples.
// Types.
// Higher order functions and immutability.
//

// in Visual Studio, select some code and press Alt + Enter to send it to F# Interactive
// you can type directly
let x = 2
let y = x * x * x



// converting between types needs to be explicit
// this doesn't work:
// let z = 2 + 3.0
let z = ((float x) ** 3.0)

// Variables are immutable by default. If we want to modify them, we have to say so in advance:
let mutable m = 20
m <- m * 2
m

// Functions.
//
let addOne x =
    x + 1

addOne 42
    
let add x y = x + y 

// We can add type annotations if the compiler does not give us what we want.
let concatenate (x: string) (y: string) = x + y
concatenate "conc" "atenate"

// Easy, now lists.
//
[1;2;3]
[1..10]
[100 .. -1 .. 1]

// arrays:
let a = [|0 .. 9|]
a.[5]
a.[5 .. 7]

// list/array comprehensions!
let squares = [for i in 1.. 10 -> i * i]

let unitMatrix = [| for i in 1..5 ->
                    [| for j in 1..5 -> if i = j then 1 else 0 |]
                 |]       


// sequences: lazily evaluated one by one element
let s = {1 .. 10}

// useful built-in functions
List.length squares
List.head squares
List.append squares squares

// We can convert between lists/arrays/sequences
Array.toList [|1 .. 10|]
Seq.ofList [1 .. 10]
List.ofSeq {1 .. 10}
Seq.toArray {1 .. 10}
// etc.

// Pipes

addOne 42
42 |> addOne

squares |> List.length

squares |> List.head |> addOne |> addOne

// We can pipe from the other side as well.
List.length <| [1 .. 10 .. 99]
// .. or from both sides!
10 |> add <| 10

// Higher order functions and currying.
//
let applyFive f = f 5
applyFive addOne

List.map addOne [1;2;3]
[1; 2; 3] |> List.map addOne

List.sum [1;2;3;4]

let listProduct = List.reduce (*)
listProduct [1;2;3;4]

let isEven x = x % 2 = 0
[1..10] |> List.filter isEven


// Anonymous functions
//
2 |> fun x -> x + 1

[1..10] |> List.filter (fun x -> x % 2 = 0)

// No NULLs by default - but we have an option type
//
None
Some(0)
Some(0).Value

let sqrt x = 
    if x < 0.0 then None else Some(System.Math.Sqrt (x))

625.0 |> sqrt |> (fun option -> option.Value)
-1.0 |> sqrt

// Tuples
//
let t = (1,2,3)
let tt = (0, "something", [10 .. -1 .. 1])

let pointDistance point = System.Math.Sqrt ((fst point * fst point) + (snd point * snd point))
pointDistance (3., 4.)

let pointDistance2 (x,y) = System.Math.Sqrt (x * x + y * y)
pointDistance2 (3., 4.)

let myFun (x, _) = x + 1


// Finally, we can have some fun with sequences.
//
let infiniteSequence = Seq.initInfinite (fun i -> i * i)
infiniteSequence 
    |> Seq.takeWhile (fun n -> n < 1000) 
    |> Seq.toList

// https://msdn.microsoft.com/en-us/library/ee340363.aspx
let fibonacci = 
    (1, 1) 
    |> Seq.unfold (fun (prev1, prev2) -> Some((prev1 + prev2), (prev1 + prev2, prev1))) 

fibonacci
    |> Seq.skipWhile (fun n -> n < 12)
    |> Seq.take 30
    |> Seq.toArray
